﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace SimpleWars 
{
    /// <summary>
    /// 移動可能範囲を概算しておくクラス．
    /// 具体的には，以下の２つの配列を計算・保存しておく． 
    /// rangeMap[,][,]  => mapの2点間の必要移動コスト
    /// enemyMastery[,] => mapの1点がどの敵からの射程であるのかがbitフラグで入ってる
    /// </summary>
    class SAT_RangeController {



        public static int[,][,] rangeMap = null;
        // ↑↑↑
        //【rangeMap】は，ある地点から見ての，マップの各点への移動コスト余裕を見る．※値「0」で攻撃だけ可能な点
        //1次元MAPで左端(1,0)に戦車がいれば， [0, 7, 6, 5, 4, 3, 2, 1, 0, -1, -1].　※左端右端はマップのエッジ
        //地形が森だけの1次元MAPなら， 　[0, 7, 5, 3, 1, 0,-1,-1,-1, -1, -1].
        //
        //こんな風にマップのある座標（x,y）を始点とした「戦車の」移動コスト余裕が2次元配列の形で格納される，
        //2次元*2次元な配列． （ちなみに戦車である理由は，飛行機やキャノンにこの種のマップが必要なく，対空戦車は戦車と
        //　同じ移動力なためである．）


        public static int[,] enemyMastery = null;
        //【enemyMastery】は敵ユニット射程圏内の重なりパターンをビットフラグ（int）で表す
        //（例） 敵ユニットe0, e1が共に射程距離2の時，それぞれの射程内に 0001 と　0010　のビットをかけて 
        // [e0] [ ] [ ]        [ *  ] [0001] [0001]
        // [  ] [ ] [ ]    =>  [0011] [0011] [0000]
        // [e1] [ ] [ ]        [ *  ] [0010] [0010]　となる．
        // （0000）はe0からもe1からも攻撃されない場所．(0001)はe0から攻撃されうる場所．(0011)はe0とe1両方から攻撃されうる場所

        public static void initRangeMap(int[,] fieldType) {
            rangeMap = new int[fieldType.GetLength(0), fieldType.GetLength(1)][,];
            for (int y = 1; y <= fieldType.GetLength(1) - 2; y++) {
                for (int x = 1; x <= fieldType.GetLength(0) - 2; x++) {
                    initSingleRangeMap(x, y, fieldType);
                }
            }
        }



        //移動コスト配列初期化．
        private static void initSingleRangeMap(int x, int y, int[,] fieldType) {
            int[] DX = new int[4] { +1, 0, -1, 0 };
            int[] DY = new int[4] { 0, -1, 0, +1 };
            int[,] aRangeMap = new int[fieldType.GetLength(0), fieldType.GetLength(1)];
            int step = 7;//Magic Number.
            for (int xx = 0; xx <= fieldType.GetLength(0) - 1; xx++) {
                for (int yy = 0; yy <= fieldType.GetLength(1) - 1; yy++) {
                    aRangeMap[xx, yy] = -1;
                }
            }

            int[] posCounts = new int[step + 1];  // 残り移動力 index になる移動可能箇所数．0で初期化される
            int[,] posX = new int[step + 1, 64]; // 残り移動力 index になる移動可能箇所． 100は適当な数，本来 Const化すべき
            int[,] posY = new int[step + 1, 64];

            posCounts[step] = 1;
            posX[step, 0] = x;
            posY[step, 0] = y;

            aRangeMap[x, y] = step;

            // 残り移動力 restStep になる移動可能箇所から，その上下左右を見て，移動できるならリストに追加していく
            for (int restStep = step; restStep > 0; restStep--) { // 現在の残り移動コスト
                for (int i = 0; i < posCounts[restStep]; i++) {
                    int xlook = posX[restStep, i];  // 注目する場所
                    int ylook = posY[restStep, i];

                    // この (x,y) の上下左右を見る
                    for (int r = 0; r <= 3; r++) {
                        int newx = xlook + DX[r]; // 上下左右
                        int newy = ylook + DY[r];

                        if (newx == 0 || newx == aRangeMap.GetLength(0) - 1 ||
                            newy == 0 || newy == aRangeMap.GetLength(1) - 1) { continue; }

                        int newrest = restStep - Spec.moveCost[2, fieldType[newx, newy]]; // 移動後の残り移動コスト
                        //負なら，0を入れて終わり（単なる攻撃可能範囲．）
                        if (newrest <= 0 && aRangeMap[newx, newy] < 0) {
                            aRangeMap[newx, newy] = 0;
                            continue;  // 周囲にあたったか，移動力が足りない →進入不可
                        }

                        // すでに移動可能マークがついた場所じゃなければ，移動可能箇所に追加する
                        if (newrest > aRangeMap[newx, newy]) {
                            aRangeMap[newx, newy] = newrest;
                            posX[newrest, posCounts[newrest]] = newx;
                            posY[newrest, posCounts[newrest]] = newy;
                            posCounts[newrest]++;
                        }
                    }
                }
            }
            rangeMap[x, y] = aRangeMap;
        }


        //敵射程の重なりを表す配列の初期化
        //※注　initRangeMapの前に呼んでしまうとバグる
        public static void initEnemyMastery(Map map, int teamColor) {
            int[,] fieldType = map.getFieldTypeArray();
            enemyMastery = new int[fieldType.GetLength(0), fieldType.GetLength(1)];

            List<Unit> enemyUnits = map.getUnitsList(teamColor, false, false, true);
            int base2 = 1;
            foreach (Unit eneU in enemyUnits) {
                int eneX = eneU.getXpos();
                int eneY = eneU.getYpos();

                //(遠距離ユニット）
                if (!eneU.getSpec().isDirectAttackType()) {
                    for (int x = 1; x < fieldType.GetLength(0) - 1; x++) {
                        for (int y = 1; y < fieldType.GetLength(1) - 1; y++) {
                            int dist = Math.Abs(eneU.getXpos() - x) + Math.Abs(eneU.getYpos() - y);
                            if (2 <= dist && dist <= 3) {
                                //攻撃が届く圏内
                                enemyMastery[x, y] = enemyMastery[x, y] + base2;
                            }
                        }
                    }
                } //(飛行系ユニットか歩兵)
                else if (eneU.getSpec().isAirUnit() || eneU.getSpec().getUnitStep() == 3) {
                    for (int x = 1; x < fieldType.GetLength(0)-1; x++) {
                        for (int y = 1; y < fieldType.GetLength(1)-1; y++) {
                            int dist = Math.Abs(eneU.getXpos() - x) + Math.Abs(eneU.getYpos() - y);
                            if (dist <= eneU.getSpec().getUnitStep() + 1) {
                                //攻撃が届く圏内
                                enemyMastery[x, y] = enemyMastery[x, y] + base2;
                            }
                        }
                    }
                } else {
                    //非飛行系かつ非歩兵
                    for (int x = 1; x < fieldType.GetLength(0)-1; x++) {
                        for (int y = 1; y < fieldType.GetLength(1)-1; y++) {
                            if (rangeMap[eneX, eneY][x, y] >= 0) {
                                //攻撃が届く圏内
                                enemyMastery[x, y] = enemyMastery[x, y] + base2;
                            }
                        }
                    }
                }
                base2 *= 2; //←この処理で何をやりたいのかは「C#　ビットフラグ」などを検索参照            }
            }
        }

        
    }

}
